/**********************************************************************************************************************
 Copyright (c) 2017 Roel Arits

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 **********************************************************************************************************************/

// The software uses the TimerOne library to generate a PWM signal on an output pin 10 of an 
// Arduino Pro Mini or Arduino Uno. This PWM signal is used to generate audio frequency tones.
// There are 2 tunes available (Jingle Bells and Mission Impossible) that can be selected by making 
// a digital input either high (not connected, internal pullup pulls input high) or low (connected to ground).
// The tune note frequencies, duration and pauzes are all defined in arrays (lookup tables).
// The tunes can be played using a loudspeaker connected to the PWM output via a electrolytic capacitor to block DC
// or via a high voltage flyback transformer that is controlled by a MOSFET. 

// When using a loudspeaker, the duty cycle for the PWM output should be set to 50% and the PWM period can be rather 
// short (higher frequencies). So for a loudspeaker the tune note frequencies can be multiplied by a fixed factor. 

// When usng a flyback transformer, the duty cycle should be low (10%) because the current through the flyback transformer
// is directly proportional to the duty cycle of the PWM signal. The note frequencies should preferably be low (longer PWM
// periods), because the lower frequencies are much better reproduced by the sparks. Higher frequencies will have much less
// volume. So the frequencies of the notes can be divided by a fixed factor.
// When you barely hear a sound and the sparks are too small, then try to increase the dutycycle with steps of 5%.
// sound is good, but the flyback driver MOSFET gets hot and the current consumption is high, then decrease the dutycycle.

/////////////////////////////////////////////////////////////
// !! BE VERY CAREFULL WHEN USING A FLYBACK TRANSFORMER !! //
/////////////////////////////////////////////////////////////

/***********************************************************************************************************************
 * I N C L U D E S
 **********************************************************************************************************************/
#include "TimerOne.h"

/***********************************************************************************************************************
 * L O C A L   S Y M B O L   D E F I N I T I O N S (defines, typedefs)
 **********************************************************************************************************************/

// Definitions for the state machine that will play the notes
typedef enum
{
  TUNE_STATE_PLAY_NOTE,
  TUNE_STATE_SUSTAIN_NOTE,
  TUNE_STATE_STOP_NOTE,
  TUNE_STATE_WAIT,
} TTuneState;

// All global variables related to the tune playing collected in 1 structure
typedef struct
{
	uint16_t    noteLengthCount;
	uint16_t    pauzeCount;

	uint8_t     noteIndex;
	TTuneState  state;

	//bool        play; // not needed
	bool        init;
} TTuneGlobalVars;

// Tune pointers to browse through the lookup tables for the note frequencies, duration and pause
typedef struct
{
  float     *pNoteFrequencies;
  uint16_t  *pNoteDurations; // in counts of 10ms
  uint16_t  *pNotePauzes; // in counts of 10ms
  uint16_t  numNotes; // amount of notes in the selected tune
} TTune;

// Output pin for the PWM audio signal
#define TUNE_AUDIO_OUT_PIN         10

// Input pin to select which tune to play : 
// Jingle bells is the default that is played when the pin is left open
// Mission impossible is played when the pin is connected to Vcc
#define TUNE_SELECT_INPUT_PIN       9

//============================================================================================================================
//MISION IMPOSSIBLE TUNE

// Here the parts of the tune are defined as notes with the corresponding note frequency, note duration and pause after the note.
// This is all done manually by first finding a midi file or a tab file of the tune on the internet and then extracting the notes
// and converting them to frequency valus in Hz. The note duration and pauses were defined by trial and error.

#define NUM_NOTES_TUNE1      70

                             //A     A       C       D        A       A       G       G# 
#define TUNE1_A_NOTES        27.5,   27.5,   32.7,   36.71,   27.5,   27.5,   24.5,   25.96,
#define TUNE1_A_DUR          38,     38,     28,     28,      38,     38,     28,     28,
#define TUNE1_A_PAUSE        2,      4,      2,      2,       2,      4,      2,      2,

                             //D     D       F       G        D       D       C       C# 
#define TUNE1_B1_NOTES       18.35,  18.35,  21.83,  24.50,   18.35,  18.35,  16.35,  17.32,
                             //D     D       F       G        D       D       G       G# 
#define TUNE1_B2_NOTES       18.35,  18.35,  21.83,  24.50,   18.35,  18.35,  24.50,  25.96,

#define TUNE1_B_LOW_DUR      38,     38,     28,     28,      38,     38,     28,     28,
#define TUNE1_B_LOW_PAUSE    2,      4,      2,      2,       2,      4,      2,      2,

                             //B      G#       D#   /   B        G#       D   /    B       G#      C#    /  B       C#   
#define TUNE1_C_NOTES        123.47,  103.83,  77.78,   123.47,  103.83,  73.42,   123.47, 103.83, 69.30,   61.74,  69.30,  
#define TUNE1_C_DUR          12,      10,      140,     12,      10,      140,     12,     10,     140,     10,     10,
#define TUNE1_C_PAUSE        1,       1,       1,       1,       1,       1,       1,      1,      1,       1,      140,  

                             //B      G#       G#     / B        G#       G    /   B       G#      F#    /  F       E
#define TUNE1_D_NOTES        61.74,   51.91,   103.83,  61.74,   51.91,   98.00,   61.74,  51.91,  92.50,   87.31,  82.41  
#define TUNE1_D_DUR          12,      10,      140,     12,      10,      140,     12,     10,     140,     10,     10
#define TUNE1_D_PAUSE        1,       1,       1,       1,       1,        1,      1,      1,      1,       1,      140   
//============================================================================================================================
// JINGLE BELLS TUNE

// Here the parts of the tune are defined as notes with the corresponding note frequency, note duration and pause after the note.
// This is all done manually by first finding a midi file or a tab file of the tune on the internet and then extracting the notes
// and converting them to frequency valus in Hz. The note duration and pauses were defined by trial and error.

#define NUM_NOTES_TUNE2      102

                             //D     D       B      A      G      D
#define TUNE2_A_NOTES        18.35,  18.35,  30.87, 27.50, 24.50, 18.35,
#define TUNE2_A_DUR          20,     20,     20,    20,    20,    60,
#define TUNE2_A_PAUSE        1,      2,      1,     1,     1,     10,

                             //D     D       B      A      G       E
#define TUNE2_B_NOTES        18.35,  18.35,  30.87, 27.50, 24.50, 20.60,
#define TUNE2_B_DUR          20,     20,     20,    20,    20,    60,
#define TUNE2_B_PAUSE        1,      2,      1,     1,     1,     10,

                             //E     E       C      B      A      D
#define TUNE2_C_NOTES        20.60,  20.60,  32.70, 30.87, 27.50, 36.71,
#define TUNE2_C_DUR          20,     20,     20,    20,    20,    60,
#define TUNE2_C_PAUSE        1,      2,      1,     1,     1,     10,

                             //D     E       D      C      A       B
#define TUNE2_D_NOTES        36.71,  41.20,  36.71, 32.70, 27.50, 30.87,
#define TUNE2_D_DUR          20,     20,     20,    20,    20,    60,
#define TUNE2_D_PAUSE        1,      2,      1,     1,     1,     10,

                             //E     E       C      B      A      D       D      D      D      E      D      C      A      G      D
#define TUNE2_E_NOTES        20.60,  20.60,  32.70, 30.87, 27.50, 36.71,  36.71, 36.71, 36.71, 41.20, 36.71, 32.70, 27.50, 24.50, 36.71,
#define TUNE2_E_DUR          20,     20,     20,    20,    20,    20,     20,    20,    20,    20,    20,    20,    20,    40,    52,
#define TUNE2_E_PAUSE        1,      2,      1,     1,     1,     1,      1,     1,     1,     1,     1,     1,     1,     6,     1,

// Jingle bells refrein
                             //B    B      B      B      B      B       
#define TUNE2_F_NOTES        30.87, 30.87, 30.87, 30.87, 30.87, 30.87,
#define TUNE2_F_DUR          20,    20,    40,    20,    20,    40,
#define TUNE2_F_PAUSE        1,     1,     1,     1,     1,     1,

                             //B    D      G      A      B       
#define TUNE2_G_NOTES        30.87, 36.71, 24.50, 27.50, 30.87,
#define TUNE2_G_DUR          20,    20,    30,    10,    40,
#define TUNE2_G_PAUSE        1,     1,     1,     1,     40, 

                             //C    C      C      C      C      B      B      B      B      B      A      A      B      A      D      
#define TUNE2_H_NOTES        32.70, 32.70, 32.70, 32.70, 32.70, 30.87, 30.87, 30.87, 30.87, 30.87, 27.50, 27.50, 30.87, 27.50, 36.71,
#define TUNE2_H_DUR          20,    20,    30,    10,    20,    20,    20,    10,    10,    20,    20,    20,    20,    40,    52,
#define TUNE2_H_PAUSE        1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,

                             //C    C      C      C      C      B      B      B      B      D      D      C      A      G      
#define TUNE2_I_NOTES        32.70, 32.70, 32.70, 32.70, 32.70, 30.87, 30.87, 30.87, 30.87, 36.71, 36.71, 32.70, 27.50, 24.50,
#define TUNE2_I_DUR          20,    20,    30,    10,    20,    20,    20,    10,    10,    20,    20,    20,    20,    40,
#define TUNE2_I_PAUSE        1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     40,

//============================================================================================================================

/***********************************************************************************************************************
 * D A T A   D E C L A R A T I O N S (exported, local)           
 **********************************************************************************************************************/
//MISSION IMPOSSIBLE TUNE
float    Tune1Notes[NUM_NOTES_TUNE1] =        { TUNE1_A_NOTES  TUNE1_A_NOTES  TUNE1_B1_NOTES     TUNE1_B2_NOTES     TUNE1_A_NOTES  TUNE1_A_NOTES TUNE1_C_NOTES  TUNE1_D_NOTES };
uint16_t Tune1NoteDuration[NUM_NOTES_TUNE1] = { TUNE1_A_DUR    TUNE1_A_DUR    TUNE1_B_LOW_DUR    TUNE1_B_LOW_DUR    TUNE1_A_DUR    TUNE1_A_DUR   TUNE1_C_DUR    TUNE1_D_DUR   };
uint16_t Tune1PostNotePauze[NUM_NOTES_TUNE1]= { TUNE1_A_PAUSE  TUNE1_A_PAUSE  TUNE1_B_LOW_PAUSE  TUNE1_B_LOW_PAUSE  TUNE1_A_PAUSE  TUNE1_A_PAUSE TUNE1_C_PAUSE  TUNE1_D_PAUSE };

//JINGLE BELLS TUNE
float    Tune2Notes[NUM_NOTES_TUNE2] =        { TUNE2_A_NOTES   TUNE2_B_NOTES  TUNE2_C_NOTES   TUNE2_D_NOTES  TUNE2_A_NOTES   TUNE2_B_NOTES  TUNE2_E_NOTES  TUNE2_F_NOTES TUNE2_G_NOTES TUNE2_H_NOTES  TUNE2_F_NOTES TUNE2_G_NOTES TUNE2_I_NOTES };
uint16_t Tune2NoteDuration[NUM_NOTES_TUNE2] = { TUNE2_A_DUR     TUNE2_B_DUR    TUNE2_C_DUR     TUNE2_D_DUR    TUNE2_A_DUR     TUNE2_B_DUR    TUNE2_E_DUR    TUNE2_F_DUR   TUNE2_G_DUR   TUNE2_H_DUR    TUNE2_F_DUR   TUNE2_G_DUR   TUNE2_I_DUR };
uint16_t Tune2PostNotePauze[NUM_NOTES_TUNE2]= { TUNE2_A_PAUSE   TUNE2_B_PAUSE  TUNE2_C_PAUSE   TUNE2_D_PAUSE  TUNE2_A_PAUSE   TUNE2_B_PAUSE  TUNE2_E_PAUSE  TUNE2_F_PAUSE TUNE2_G_PAUSE TUNE2_H_PAUSE  TUNE2_F_PAUSE TUNE2_G_PAUSE TUNE2_I_PAUSE };

// Instance of the TimerOne class defined in the TimerOne library (PWM generation)
TimerOne          timer1;

TTuneGlobalVars   tuneGlobalVars;
TTune             tune;

bool              inputPinState;
bool              prevInputPinState;

/***********************************************************************************************************************
 * L O C A L   F U N C T I O N S                                   
 **********************************************************************************************************************/
//-----------------------------------------------------------------------------------------------------------------------
// L O C A L   F U N C T I O N S                                   
//-----------------------------------------------------------------------------------------------------------------------
void setup() 
{
	tuneGlobalVars.init = true;
	
	// Define the tune select pin as an input with the internal pull up activated so the pin is high when left open.
	pinMode(9, INPUT_PULLUP);

	// put your setup code here, to run once:
	timer1.initialize();
}

//-----------------------------------------------------------------------------------------------------------------------
// L O C A L   F U N C T I O N S                                   
//-----------------------------------------------------------------------------------------------------------------------
void loop() 
{
  // put your main code here, to run repeatedly:

  // Get input state
  // The internal pullup of the input is set, so when the input is open, it will be high
  inputPinState = (digitalRead(TUNE_SELECT_INPUT_PIN) == LOW);

  // When input is low,  play tune 1 (mission impossible)
  // When input is high, play tune 2 (jingle bells)
  if (inputPinState)
  {
    // MISSION IMPOSSIBLE (tune 1)
    tune.pNoteFrequencies = Tune1Notes;
    tune.pNoteDurations = Tune1NoteDuration;
    tune.pNotePauzes = Tune1PostNotePauze;
    tune.numNotes = NUM_NOTES_TUNE1;
  }
  else
  {
    // JINGLE BELLS (tune 2)
    tune.pNoteFrequencies = Tune2Notes;
    tune.pNoteDurations = Tune2NoteDuration;
    tune.pNotePauzes = Tune2PostNotePauze;
    tune.numNotes = NUM_NOTES_TUNE2;
  }
  
  // When changing to the other tune, then re-initialise all the parameters
  if (inputPinState != prevInputPinState)
  {
      tuneGlobalVars.init = true;
  }
  
  // Check if we need to initialise all parameters
  if (tuneGlobalVars.init)
  {
    tuneGlobalVars.init = false;
    tuneGlobalVars.noteLengthCount = 0;
    tuneGlobalVars.pauzeCount = 0;
    tuneGlobalVars.noteIndex = 0;
    //tuneGlobalVars.play = true;
  }

  // We use a 10ms timing loop created by a delay.
  // Because this application only needs to play a tune and has no other tasks, using a simple delay is good enough
  // because there are no other tasks running. Otherwise we'd need a timer interrupt to create a stable timing.
  delay(10);

  // Check if we need to play
  //if (tuneGlobalVars.play)
  //{
    // Check if we need to play the note or sustain a note for a certain duration or if we are waiting/pausing before playing the next note
    switch (tuneGlobalVars.state)
    {
      case TUNE_STATE_PLAY_NOTE:
      {
          // Get the current note from the array and convert the note frequency to a period expressed in microseconds.
          // Don't make the dutycycle too big, because this will increase the current through the flyback transformer
          timer1.pwm(TUNE_AUDIO_OUT_PIN, 10, 1000000.0 / (*(tune.pNoteFrequencies + tuneGlobalVars.noteIndex) * 3L));

          // Sustain the note that is being played.
          tuneGlobalVars.state = TUNE_STATE_SUSTAIN_NOTE;
      }
      break;
      
      case TUNE_STATE_SUSTAIN_NOTE:
      {
          // Get the duration of the current note and sustain the note for this duration
          if (tuneGlobalVars.noteLengthCount > *(tune.pNoteDurations + tuneGlobalVars.noteIndex))
          {
              tuneGlobalVars.noteLengthCount = 0;
          
              // Go to the next state to stop the note
              tuneGlobalVars.state = TUNE_STATE_STOP_NOTE;
          }
          else
          {
              tuneGlobalVars.noteLengthCount++;
          }
      }
      break;

      case TUNE_STATE_STOP_NOTE:
      {
          // To stop playing a note, we set the PWM period to 1us, because setting th period to 0 is not allowed
          // Don't make the dutycycle too big, because this will increase the current through the flyback transformer
          timer1.pwm(TUNE_AUDIO_OUT_PIN, 10, 1);
          
          // Go the next state to wait before playing the next note (pause time between the notes)
          tuneGlobalVars.state = TUNE_STATE_WAIT;
      }
      break;
        
      case TUNE_STATE_WAIT:
      {
        // Pause after playing the current note and before playing the next note.
        if (tuneGlobalVars.pauzeCount > *(tune.pNotePauzes + tuneGlobalVars.noteIndex))
        {
            tuneGlobalVars.pauzeCount = 0;
            tuneGlobalVars.state = TUNE_STATE_PLAY_NOTE;
            tuneGlobalVars.noteIndex++;
            
            if (tuneGlobalVars.noteIndex >= tune.numNotes)
            {
                tuneGlobalVars.noteIndex = 0;
            }
        }
        else
        {
            tuneGlobalVars.pauzeCount++;
        }
      }
      break;
    }
  //}    
  prevInputPinState = inputPinState;
}
